#include "level.h"
#include "game.h"

Level::Level() {
	this->clear();
}

void Level::clear() {
	for (int x=0;x<11;x++) {
		for (int y=0;y<11;y++) {
			this->data[y][x]=0;
		}
	}
}

/* These random functions were taken from OpenBSD's random implementation */

#define TYPE_3          3               /* x**31 + x**3 + 1 */
#define BREAK_3         128
#define DEG_3           31
#define SEP_3           3

static int32_t randtbl[DEG_3 + 1] = {
	TYPE_3,
	0x991539b1, 0x16a5bce3, 0x6774a4cd, 0x3e01511e, 0x4e508aaa, 0x61048c05,
	0xf5500617, 0x846b7115, 0x6a19892c, 0x896a97af, 0xdb48f936, 0x14898454,
	0x37ffd106, 0xb58bff9c, 0x59e17104, 0xcf918a49, 0x09378c83, 0x52c7a471,
	0x8d293ea9, 0x1f4fc301, 0xc3db71be, 0x39b44e1c, 0xf8a44ef9, 0x4c8b80b1,
	0x19edc328, 0x87bf4bdd, 0xc9b240e5, 0xe9ee4b1b, 0x4382aee7, 0x535b6b41,
	0xf3bec5da,
};

static int32_t *fptr = &randtbl[SEP_3 + 1];
static int32_t *rptr = &randtbl[1];

static int32_t *state = &randtbl[1];
static int32_t *end_ptr = &randtbl[DEG_3 + 1];

static int
rando(void)
{
	int32_t i;

	*fptr += *rptr;
	i = (*fptr >> 1) & 0x7fffffff;  /* chucking least random bit */
	if (++fptr >= end_ptr) {
		fptr = state;
		++rptr;
	} else if (++rptr >= end_ptr)
		rptr = state;
	return(i);
}


static void
srando(unsigned int x)
{
	int i;
	int32_t test;
	div_t val;

	if(!x)
		x = time(NULL);
	state[0] = x;
	for (i = 1; i < DEG_3; i++) {
		val = div(state[i-1], 127773);
		test = 16807 * val.rem - 2836 * val.quot;
		state[i] = test + (test < 0 ? 2147483647 : 0);
	}
	fptr = &state[SEP_3];
	rptr = &state[0];
	for (i = 0; i < 10 * DEG_3; i++)
		(void)rando();
}

/* The following parameters might need tuning! */
#define TIMEPERMOVE .6	/* The amount of time (measured in seconds) a player has to clear ony block in a random map */
#define MIN_LEN 20	/* The minimum number of blocks a player has to clear */
#define MAX_LEN 80	/* The maximum number of blocks a player has to clear */
#define WALKALONG 70	/* The probability in % the random map will try to walk along the line already walking.  Shouldn't be < 25 */

#define UP 0
#define DOWN 1
#define LEFT 2
#define RIGHT 3
int Level::random(unsigned int seed, int severity)
{
	int x, y, dir, i, len;
	int valid[4], sum;

	srando(seed);

	this->clear();

	len = MIN_LEN + rando()%(MAX_LEN - MIN_LEN + 1);


	x = this->startX = rando()%11;
	y = this->startY = rando()%11;

	this->data[y][x] = 1;
	
	dir = rando()%4;

	for(i=0; i<len; i++) {
		valid[0] = valid[1] = valid[2] = valid[3] = 0;
		if(x>0 && this->data[y][x-1] != 5)
			valid[LEFT] = 1;
		if(x<10 && this->data[y][x+1] != 5)
			valid[RIGHT] = 1;
		if(y>0 && this->data[y-1][x] != 5)
			valid[UP] = 1;
		if(y<10 && this->data[y+1][x] != 5)
			valid[DOWN] = 1;

		sum = valid[0] + valid[1] + valid[2] + valid[3];

		if(sum == 0) { // Nowhere left to worm, so end
			this->data[y][x] = SYMBOL_EXIT;
			this->timeout = (int ((double)i * TIMEPERMOVE)) * (4 - severity);
			return 1;
		}
		if(!valid[dir] || rando()%100 >= WALKALONG) // Change direction
			do {
				dir = rando()%4;
			} while(!valid[dir]);
		switch(dir) {
			case UP: y--;
				 break;
			case DOWN: y++;
				   break;
			case RIGHT: x++;
				    break;
			case LEFT: x--;
				   break;
		}
		this->data[y][x]++;
	}

	this->timeout = (int ((double)len * TIMEPERMOVE)) * (4 - severity);

	this->data[y][x] = SYMBOL_EXIT;
	return 1;
}



int Level::load(int level, int severity) {
	string row;

	this->clear();

	string filename;

	char buf[10];

	snprintf(buf,10,"%03d",level);

	filename=SHAREDIR"/data/levels/level";
	filename+=buf;
	filename+=".txt";

	ifstream fin(filename.c_str());
	int rowcount=0;
	int value;

	if (!fin) return 0;
	while (fin >> row) {
		switch (rowcount) {
			case 0:
				sscanf(row.c_str(),"%d,%d,%d",&this->startX,&this->startY,&this->timeout);
				this->timeout*=4-severity;
				break;
			default:
				if (row.length()>10) {
					for (int i=0;i<11;i++) {
						value=row[i]-0x30;
						if (value>-1) {
							this->data[rowcount-1][i]=value;
						}
						// cout << value << " ";
					}
				}
				break;
		}
		// cout << endl;
		rowcount++;
	}
	fin.close();
	return 1;
}

int Level::isCleared() {
	for (int x=0;x<11;x++) {
		for (int y=0;y<11;y++) {
			if (
					(this->data[y][x]<6) &&
					(this->data[y][x]>0)
			) {
				return 0;
			}
		}
	}
	return 1;
}

int Level::getTimeout() {
	return this->timeout;
}

int Level::getStartX() {
	return this->startX;
}

int Level::getStartY() {
	return this->startY;
}

